import re
from logging import Logger
from typing import List, Dict, Optional
from saviialib.general_types.error_types.api.epii_api_error_types import (
    BackupSourcePathError,
)
from saviialib.libs.directory_client import DirectoryClient, DirectoryClientArgs

dir_client = DirectoryClient(DirectoryClientArgs(client_name="os_client"))


def extract_error_information(error: str) -> Optional[Dict[str, str]]:
    match = re.search(r"(\d+), message='([^']*)', url=\"([^\"]*)\"", error)
    if match:
        return {
            "status_code": match.group(1),
            "message": match.group(2),
            "url": match.group(3),
        }
    return None


def explain_status_code(status_code: int) -> str:
    explanations = {
        404: "Probably an error with file or folder source path.",
        403: "Permission denied when accessing the source path.",
        500: "Internal server error occurred during upload.",
    }
    return explanations.get(status_code, "Unknown error occurred.")


def extract_error_message(logger: Logger, results: List[Dict], success: float) -> str:
    logger.info(
        "[local_backup_lib] Not all files uploaded ⚠️\n"
        f"[local_backup_lib] Files failed to upload: {(1 - success):.2%}"
    )

    failed_files = [item for item in results if not item.get("uploaded")]

    error_data = []
    for item in failed_files:
        error_info = extract_error_information(item.get("error_message", ""))
        if error_info:
            error_data.append(
                {
                    "file_name": item["file_name"],
                    "status_code": error_info["status_code"],
                    "message": error_info["message"],
                    "url": error_info["url"],
                }
            )

    # Group errors by code.
    grouped_errors: Dict[str, List[Dict]] = {}
    for error in error_data:
        code = error["status_code"]
        grouped_errors.setdefault(code, []).append(error)

    # Summary
    for code, items in grouped_errors.items():
        logger.info(
            f"[local_backup_lib] Status code {code} - {explain_status_code(int(code))}"
        )
        for item in items:
            logger.info(
                f"[local_backup_lib] File {item['file_name']}, url: {item['url']}, message: {item['message']}"
            )

    failed_file_names = [item["file_name"] for item in failed_files]
    return f"Failed files: {', '.join(failed_file_names)}."


def parse_execute_response(results: List[Dict]) -> Dict[str, List[str]]:
    try:
        return {
            "new_files": len(
                [item["file_name"] for item in results if item.get("uploaded")]
            ),
        }
    except (IsADirectoryError, AttributeError, ConnectionError) as error:
        raise BackupSourcePathError(reason=error)


def show_upload_result(uploaded: bool, file_name: str) -> str:
    status = "✅" if uploaded else "❌"
    message = "was uploaded successfully" if uploaded else "failed to upload"
    result = f"[local_backup_lib] File {file_name} {message} {status}"
    return result


def calculate_percentage_uploaded(results: List[Dict], total_files: int) -> float:
    uploaded_count = sum(
        1 for result in results if isinstance(result, dict) and result.get("uploaded")
    )
    return (uploaded_count / total_files) * 100 if total_files > 0 else 0


async def count_files_in_directory(path: str, folder_name: str) -> int:
    return len(await dir_client.listdir(dir_client.join_paths(path, folder_name)))



# read version from installed package
from importlib.metadata import version

__version__ = version("saviialib")

from .services.epii.api import EpiiAPI
from .general_types.api.epii_api_types import EpiiAPIConfig

__all__ = ["EpiiAPI", "EpiiAPIConfig"]



from typing import Any, Dict, List

from .controllers.types.update_thies_data_types import UpdateThiesDataControllerInput
from .controllers.types.upload_backup_to_sharepoint_types import (
    UploadBackupToSharepointControllerInput,
)
from .controllers.update_thies_data import UpdateThiesDataController
from .controllers.upload_backup_to_sharepoint import UploadBackupToSharepointController
from saviialib.general_types.api.epii_api_types import (
    EpiiUpdateThiesConfig,
    EpiiSharepointBackupConfig,
    EpiiAPIConfig,
)


class EpiiAPI:
    """
    EpiiAPI is a service class that provides methods to interact with Patagonia Center system.
    """

    def __init__(self, config: EpiiAPIConfig):
        self.ftp_port = config.ftp_port
        self.ftp_host = config.ftp_host
        self.ftp_user = config.ftp_user
        self.ftp_password = config.ftp_password
        self.sharepoint_client_id = config.sharepoint_client_id
        self.sharepoint_client_secret = config.sharepoint_client_secret
        self.sharepoint_tenant_id = config.sharepoint_tenant_id
        self.sharepoint_tenant_name = config.sharepoint_tenant_name
        self.sharepoint_site_name = config.sharepoint_site_name
        self.logger = config.logger

    async def update_thies_data(
        self, sharepoint_folders_path: List[str], ftp_server_folders_path: List[str]
    ) -> Dict[str, Any]:
        """Updates data from a THIES Data Logger by connecting to an FTP server
        and transferring data to specified Sharepoint folders.

        Args:
            sharepoint_folders_path (list): List of Sharepoint folder paths for AVG and EXT data.
            The AVG path must be the first element.
            ftp_server_folders_path (list): List of FTP server folder paths for AVG and EXT data.
            The AVG path must be the first element.

        Returns:
            dict: A dictionary representation of the API response.
        """
        config = EpiiUpdateThiesConfig(
            ftp_port=self.ftp_port,
            ftp_host=self.ftp_host,
            ftp_user=self.ftp_user,
            ftp_password=self.ftp_password,
            sharepoint_client_id=self.sharepoint_client_id,
            sharepoint_client_secret=self.sharepoint_client_secret,
            sharepoint_site_name=self.sharepoint_site_name,
            sharepoint_tenant_id=self.sharepoint_tenant_id,
            sharepoint_tenant_name=self.sharepoint_tenant_name,
            logger=self.logger,
        )
        controller = UpdateThiesDataController(
            UpdateThiesDataControllerInput(
                config, sharepoint_folders_path, ftp_server_folders_path
            )
        )
        response = await controller.execute()
        return response.__dict__

    async def upload_backup_to_sharepoint(
        self, local_backup_source_path: str, sharepoint_destination_path: str
    ) -> Dict[str, Any]:
        """Migrate a backup folder from Home assistant to Sharepoint directory.
        Args:
            local_backup_source_path (str): Local path to backup.
        Returns:
            response (dict): A dictionary containing the response from the upload operation.
                This dictionary will typically include information about the success or
                failure of the upload, as well as any relevant metadata.
        """
        config = EpiiSharepointBackupConfig(
            sharepoint_client_id=self.sharepoint_client_id,
            sharepoint_client_secret=self.sharepoint_client_secret,
            sharepoint_site_name=self.sharepoint_site_name,
            sharepoint_tenant_id=self.sharepoint_tenant_id,
            sharepoint_tenant_name=self.sharepoint_tenant_name,
            logger=self.logger,
        )

        controller = UploadBackupToSharepointController(
            UploadBackupToSharepointControllerInput(
                config, local_backup_source_path, sharepoint_destination_path
            )
        )
        response = await controller.execute()
        return response.__dict__



from saviialib.libs.directory_client.directory_client_contract import (
    DirectoryClientContract,
)
import os
import asyncio


class OsClient(DirectoryClientContract):
    @staticmethod
    def join_paths(*paths: str) -> str:
        return os.path.join(*paths)

    @staticmethod
    async def path_exists(path: str) -> bool:
        return await asyncio.to_thread(os.path.exists, path)

    @staticmethod
    async def listdir(path: str) -> list:
        return await asyncio.to_thread(os.listdir, path)

    @staticmethod
    async def isdir(path: str) -> list:
        return await asyncio.to_thread(os.path.isdir, path)

    @staticmethod
    async def makedirs(path: str) -> None:
        return await asyncio.to_thread(os.makedirs, path, exist_ok=True)



from abc import ABC, abstractmethod


class DirectoryClientContract(ABC):
    @abstractmethod
    def join_paths(self, *paths: str) -> str:
        pass

    @abstractmethod
    async def path_exists(self, path: str) -> bool:
        pass

    @abstractmethod
    async def listdir(self, path: str) -> list:
        pass

    @abstractmethod
    async def isdir(self, path) -> bool:
        pass

    @abstractmethod
    async def makedirs(self, path: str) -> None:
        pass
